#pragma once

#include "Subsystems/EngineSubsystem.h"
#include "SkyPortalRunner.h"


#include "SkyPortalSubsystem.generated.h"

#pragma region Definitions

UENUM(BlueprintType)
enum EPortalSide {
	LEFT UMETA(DisplayName = "Left side"),
	RIGHT UMETA(DisplayName = "Right side"),
	BOTH UMETA(DisplayName = "Both left and right"),
	TRAP UMETA(DisplayName = "Trap")
};

UENUM(BlueprintType)
enum EPortalCommand {
	A UMETA(DisplayName = "Activate"),
	C UMETA(DisplayName = "Color"),
	J UMETA(DisplayName = "Advanced color"),
	L UMETA(DisplayName = "Trap color"),
	M UMETA(DisplayName = "Music"),
	Q UMETA(DisplayName = "Query"),
	R UMETA(DisplayName = "Ready"),
	S UMETA(DisplayName = "Status")
};








//// Delegates
DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FOnSkylanderAddedDelegate, int32, SkylanderID, int32, Index);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FOnSkylanderRemovedDelegate, int32, SkylanderID, int32, Index);

#pragma endregion








/*
* Handle all the blueprints functions and game logic.
* 
*/
UCLASS(MinimalAPI)
class USkyPortalSubsystem : public UEngineSubsystem
{
	GENERATED_BODY()

public:
	// Override initialization and deinitialization methods
	virtual void Initialize(FSubsystemCollectionBase& Collection) override;
	virtual void Deinitialize() override;

	/********* Portal Actions *************/

	/* The first function to run, before anything else.
	*  It will re-init and prepare the portal for receiving/sending inputs
	*
	*  return false if portal is not found
	*/
	UFUNCTION(BlueprintCallable, CallInEditor, meta = (Category = "SkyPortal"))
	SKYPORTAL_API bool ConnectPortal();


	UFUNCTION(BlueprintCallable, CallInEditor, meta = (Category = "SkyPortal|Commands"))
	SKYPORTAL_API void Ready();

	/*Send a **Status** command, to see if the portal is ready to receive new commands*/
	UFUNCTION(BlueprintCallable, BlueprintPure, meta = (Category = "SkyPortal|NOT FUNCTIONING"))
	SKYPORTAL_API bool bIsPortalReady();

	UFUNCTION(BlueprintCallable, meta = (Category = "SkyPortal|NOT FUNCTIONING"))
	SKYPORTAL_API void SendPortalCommand(EPortalCommand Command);

	UFUNCTION(BlueprintCallable, meta = (Category = "SkyPortal|NOT FUNCTIONING"))
	SKYPORTAL_API void SendPortalSound(USoundWave* Sound);



	/* Change portal color, ideally should be called just at the start.For gameplay usage, use ChangePortalColorside()*/
	UFUNCTION(BlueprintCallable, CallInEditor, meta = (AutoCreateRefTerm = "Color", Category = "SkyPortal|Cosmetic"))
	SKYPORTAL_API void ChangePortalColor(const FLinearColor& Color = FLinearColor::Green);

	/**
	 * Change the color of the portal, can separate side and even trap ligth
	 * @param	NextColor	New color
	 * @param	PortalSide	The actors to record
	 * @param	BlendTime	Blend between current color and NextColor, in milliseconds
	 */
	UFUNCTION(BlueprintCallable, CallInEditor, meta = (AutoCreateRefTerm = "NextColor", Category = "SkyPortal|Cosmetic", HideAlphaChannel))
	SKYPORTAL_API void ChangePortalColorside(const FLinearColor& NextColor = FLinearColor::Green, const EPortalSide PortalSide = EPortalSide::BOTH, const float BlendTime = 500.0f);


	// Blueprint-assignable event property
	UPROPERTY(BlueprintAssignable, Category = "SkyPortal|Skylander")
	FOnSkylanderAddedDelegate OnSkylanderAdded;

	// Blueprint-assignable event property
	UPROPERTY(BlueprintAssignable, Category = "SkyPortal|Skylander")
	FOnSkylanderRemovedDelegate OnSkylanderRemoved;

	UPROPERTY(BlueprintReadOnly)
	FPortalStatusData CurrentStatusData;

	UPROPERTY(BlueprintReadOnly)
	FPortalStatusData OldStatusData;

	UPROPERTY(BlueprintReadOnly)
	uint32 PortalId;

	EPortalCommand GetPortalCommandFromChar(unsigned char Char);


	UPROPERTY(VisibleAnywhere, BlueprintReadOnly)
	bool bPortalConnected = false;


	
	UFUNCTION()
	TArray<uint8> QueryBlock(uint8 characterIndex, uint8 block = 0x00);

	USkyPortalIO* PortalHandle;

private:
	FPortalStatusData ParsePortalStatus(const WriteBlock& ResponseBlock);
	



	static void Sleep(int sleepMs);



	void ActivatePortal(int active);

	


	// Block/byte related data write/read functions
	
	
	FigureDataBlock ReadFigureBlocks(uint8 FigureIndex);
	bool FalsePositive() const;

	// Pointer to the status checker thread
	FPortalStatusChecker* StatusChecker;
	FRunnableThread* StatusCheckerThread;


protected:

};


